open Core.Std
open Parser

let rec eval_expr env expr =
  let ii_op e e' f =
    match eval_expr env e, eval_expr env e' with
    | Some (IntVal n), Some (IntVal n') -> Some (IntVal (f n n'))
    | _ -> None
  in
  let ib_op e e' f =
    match eval_expr env e, eval_expr env e' with
    | Some (IntVal n), Some (IntVal n') -> Some (BoolVal (f n n'))
    | _ -> None
  in
  let bb_op e e' f =
    match eval_expr env e, eval_expr env e' with
    | Some (BoolVal n), Some (BoolVal n') -> Some (BoolVal (f n n'))
    | _ -> None
  in 
  match expr with
  | Var s -> Hashtbl.find env s
  | IntLit n -> Some (IntVal n)
  | BoolLit b -> Some (BoolVal b)
  | Neg e -> 
     (match eval_expr env e with
      | Some (IntVal n) -> Some (IntVal (-n))
      | _ -> None
     )
  | Not e ->
     (match eval_expr env e with
      | Some (BoolVal b) -> Some (BoolVal (not b))
      | _ -> None
     )
  | Plus (e, e') -> ii_op e e' (+)
  | Minus (e, e') -> ii_op e e' (-)
  | Times (e, e') -> ii_op e e' ( * )
  | Divide (e, e') -> ii_op e e' (/)
  | LT (e, e') -> ib_op e e' (<)
  | GT (e, e') -> ib_op e e' (>)
  | Equal (e, e') -> ib_op e e' (=)
  | And (e, e') -> bb_op e e' (&&)
  | Or (e, e') -> bb_op e e' (||)

let run_stmt s =
  let env = String.Table.create () in
  let rec loop s =
    match s with
    | Assign (v, e) ->
       (match eval_expr env e with
	| Some data -> Hashtbl.replace env ~key:v ~data
	| None -> 
	   failwith "Couldn't evaluate expression in assignment."
       )
    | If (cond, s, s') ->
       (match eval_expr env cond with
	| Some (BoolVal b) -> if b then loop s else loop s'
	| _ -> failwith "Expected bool in if."
       )
    | While (cond, s) ->
       (match eval_expr env cond with
	| Some (BoolVal b) -> 
	   if b then begin
	       loop s;
	       loop (While (cond, s))
	     end
	| _ -> failwith "Expected bool in while."
       )
    | Block ss -> List.iter ss ~f:loop
    | Print e ->
       (match eval_expr env e with
	| Some (IntVal n) -> printf "%d\n" n
	| Some (BoolVal b) -> printf "%b\n" b
	| None ->
	   failwith "Couldn't evaluate expression in print."
       )
    | Empty -> ()
  in
  loop s

let spec =
  let open Command.Spec in
  empty
  +> anon ("filename" %: string)
;;

let command =
  Command.basic
    ~summary:"Run a program in our fake language!"
    spec
    (fun filename () ->
     match parse_file filename with
     | None -> failwith "Parse error."
     | Some s -> run_stmt s
    )
;;

Command.run command
