open Core.Std

module Input : sig
    type t
    val of_string : string -> t
end

(** A ['result t] is a parser that consumes an [Input.t] and either produces a ['result option].
     If the output is [Some x], we say that the parser has succeeded and produced [x]. If the output
     is [None], we say that the parser has failed. *)
type 'result t = Input.t -> ('result * Input.t) option

(** [return x] is a parser that consumes no input and produces [x]. *)
val return : 'a -> 'a t

(** [x >>= f] is a parser that first performs [x] and, if it succeeds in producing some result [y],
     then performs [f y]. *)
val (>>=) : 'a t -> ('a -> 'b t) -> 'b t

(** [x <|> y] is a parser that tries [x] and, if that fails, does [y] instead. *)
val (<|>) : 'a t -> 'a t -> 'a t

(** [x >> y] is a parser that first performs [x] and, if it succeeds, ignores the result and
     performs [y]. *)
val (>>)  : _ t -> 'a t -> 'a t

(** [any] is a parser that consumes one character and produces it, or fails if the input has been exhausted. *)
val any : char t

(** [fail] is a parser that always fails, consuming nothing. *)
val fail : _ t

(** [satisfy p] is a parser that consumes a character [c] and succeeds if and only if [p c] is true. If it
     does succeed, it produces [c]. *)
val satisfy : (char -> bool) -> char t

(** [one_of cs] is a parser that consumes a character [c] and succeeds, producing [c], if and only if [c] is
     in the list [cs]. *)
val one_of : char list -> char t

(** [eof] is a parser that succeeds if and only if the input has been exhausted. *)
val eof : unit t

(** [between a b x] is a parser that first runs [a], then [x], then [b], and if they all succeeded, produces
     the output from [x]. *)
val between : _ t -> _ t -> 'a t -> 'a t

(** [many x] is a parser that runs [x] as many times as possible, producing a list of the results. It always
     succeeds: if the first attempt at running [x] fails, [many x] will produce an empty list.
    [many1 x] is like [many x], except it only succeeds if [x] is successful at least once. *)
val many  : 'a t -> 'a list t
val many1 : 'a t -> 'a list t

(** [chain x c] is a parser that consumes a sequence of things matching [x] connected by things matching [c],
     so of the form "x c x c ... c x". The parser [c] produces a function that is used to combine the two
     neighboring outputs of [x]. *)
val chain : 'a t -> ('a -> 'a -> 'a) t -> 'a t

(** These parsers all consume a single character and produce it if it is a space, a letter, a digit, an
     alphanumeric character, or a non-space non-alphanumeric character respectively. *)
val space         : char t
val letter        : char t
val digit         : char t
val alpha_num     : char t
val non_alpha_num : char t

(** [spaces] is a parser that consumes as much whitespace as possible, producing (). *)
val spaces : unit t

(** [token_of x] is a parser that runs [x] as many times as possible, combining the results into a string. Surrounding
     whitespace is ignored. *)
val token_of : char t -> string t

(** [token s] is a parser that consumes a token (i.e., a whitespace separated sequence of either all alphanumeric
     characters or all non-alphanumeric characters) and succeeds if that token was identical to [s], failing otherwise. *)
val token : string -> unit t

(** [parse p s] runs the parser [p] on the string [s], producing [Some result] if successful and [None] if not. *)
val parse : 'a t -> string -> 'a option
